"""pdval - Pandas validation using Annotated types and decorators."""

__version__ = "0.1.1"

from pdval.validation import (
  Datetime,  # Changed from DateTimeIndexed
  Finite,
  Ge,
  Gt,
  HasColumn,
  HasColumns,
  Index,  # Added
  Le,
  Lt,
  MonoDown,
  MonoUp,
  NonNaN,
  NonNegative,
  Positive,
  Validated,
  Validator,
  validated,
)

__all__ = [
  "Datetime",
  "Finite",
  "Ge",
  "Gt",
  "HasColumn",
  "HasColumns",
  "Index",
  "Le",
  "Lt",
  "MonoDown",
  "MonoUp",
  "NonNaN",
  "NonNegative",
  "Positive",
  "Validated",
  "Validator",
  "__version__",
  "validated",
]



"""Pandas validation using Annotated types and decorators (powered by pandera).

This module provides a decorator-based validation system for pandas DataFrames
and Series using Python's Annotated types, wrapping pandera for validation logic.
"""

from __future__ import annotations

# pyright: reportOperatorIssue=false, reportUnknownMemberType=false, reportUnknownVariableType=false
import functools
import inspect
import typing
from collections.abc import Callable
from typing import Annotated, Any, ParamSpec, TypeVar, get_args, get_origin

import numpy as np
import pandas as pd
import pandera.pandas as pa
from pandera.errors import SchemaError

# Validated alias for Annotated
Validated = Annotated

# Validator Classes


class Validator:
  """Base class for validators."""

  def get_checks(self) -> list[pa.Check]:
    """Return a list of pandera Checks."""
    return []

  def validate(self, data: Any) -> Any:  # noqa: ANN401
    """Validate data directly using pandera checks."""
    checks = self.get_checks()
    if not checks:
      return data

    if isinstance(data, pd.DataFrame):
      # For DataFrame, we need to know if checks are global or column-specific
      # Most simple validators (Finite, NonNaN) are element-wise or global
      # But pandera Checks on DataFrame are usually global unless specified
      # Let's assume simple validators apply to the whole dataframe
      schema = pa.DataFrameSchema(checks=checks, coerce=False)
      schema.validate(data)
    elif isinstance(data, pd.Series):
      schema = pa.SeriesSchema(checks=checks, coerce=False)
      schema.validate(data)

    return data


class Finite(Validator):
  """Validator for finite values (no Inf, no NaN)."""

  def get_checks(self) -> list[pa.Check]:
    # Use optimized numpy check
    return [pa.Check(lambda s: np.isfinite(s.values).all(), error="must be finite")]


class NonNaN(Validator):
  """Validator for non-NaN values."""

  def get_checks(self) -> list[pa.Check]:
    return [pa.Check(lambda s: pd.notna(s).all(), error="must not contain NaN")]


class NonNegative(Validator):
  """Validator for non-negative values (>= 0)."""

  def get_checks(self) -> list[pa.Check]:
    return [pa.Check.ge(0)]


class Positive(Validator):
  """Validator for positive values (> 0)."""

  def get_checks(self) -> list[pa.Check]:
    return [pa.Check.gt(0)]


class Datetime(Validator):
  """Validator for datetime index or values."""

  # Datetime validation is usually handled by schema dtype, but we can add a check
  def get_checks(self) -> list[pa.Check]:
    return []  # Handled by coerce/dtype in schema construction if possible


class HasColumns(Validator):
  """Validator for presence of specific columns in DataFrame."""

  def __init__(self, columns: list[str]) -> None:
    self.columns = columns

  def __class_getitem__(cls, items: str | tuple[str, ...]) -> HasColumns:
    if isinstance(items, str):
      items = (items,)
    return cls(list(items))

  def validate(self, data: Any) -> Any:  # noqa: ANN401
    if not isinstance(data, pd.DataFrame):
      raise TypeError("HasColumns validator requires a pandas DataFrame")

    missing = [col for col in self.columns if col not in data.columns]
    if missing:
      # Raise SchemaError to match pandera behavior
      raise SchemaError(schema=None, data=data, message=f"Missing columns: {missing}")
    return data


class Ge(Validator):
  """Validator that Col1 >= Col2."""

  def __init__(self, col1: str, col2: str) -> None:
    self.col1 = col1
    self.col2 = col2

  def __class_getitem__(cls, items: tuple[str, str]) -> Ge:
    return cls(items[0], items[1])

  def get_checks(self) -> list[pa.Check]:
    return [
      pa.Check(
        lambda df: (
          np.all(df[self.col1].values >= df[self.col2].values)
          if isinstance(df, pd.DataFrame)
          and self.col1 in df.columns
          and self.col2 in df.columns
          else True
        ),
        name=f"{self.col1} >= {self.col2}",
        error=f"{self.col1} must be >= {self.col2}",
      )
    ]

  def validate(self, data: Any) -> Any:  # noqa: ANN401
    if not isinstance(data, pd.DataFrame):
      raise TypeError("Ge validator requires a pandas DataFrame")
    return super().validate(data)


class Le(Validator):
  """Validator that Col1 <= Col2."""

  def __init__(self, col1: str, col2: str) -> None:
    self.col1 = col1
    self.col2 = col2

  def __class_getitem__(cls, items: tuple[str, str]) -> Le:
    return cls(items[0], items[1])

  def get_checks(self) -> list[pa.Check]:
    return [
      pa.Check(
        lambda df: (
          np.all(df[self.col1].values <= df[self.col2].values)
          if isinstance(df, pd.DataFrame)
          and self.col1 in df.columns
          and self.col2 in df.columns
          else True
        ),
        name=f"{self.col1} <= {self.col2}",
        error=f"{self.col1} must be <= {self.col2}",
      )
    ]

  def validate(self, data: Any) -> Any:  # noqa: ANN401
    if not isinstance(data, pd.DataFrame):
      raise TypeError("Le validator requires a pandas DataFrame")
    return super().validate(data)


class Gt(Validator):
  """Validator that Col1 > Col2."""

  def __init__(self, col1: str, col2: str) -> None:
    self.col1 = col1
    self.col2 = col2

  def __class_getitem__(cls, items: tuple[str, str]) -> Gt:
    return cls(items[0], items[1])

  def get_checks(self) -> list[pa.Check]:
    return [
      pa.Check(
        lambda df: (
          np.all(df[self.col1].values > df[self.col2].values)
          if isinstance(df, pd.DataFrame)
          and self.col1 in df.columns
          and self.col2 in df.columns
          else True
        ),
        name=f"{self.col1} > {self.col2}",
        error=f"{self.col1} must be > {self.col2}",
      )
    ]

  def validate(self, data: Any) -> Any:  # noqa: ANN401
    if not isinstance(data, pd.DataFrame):
      raise TypeError("Gt validator requires a pandas DataFrame")
    return super().validate(data)


class Lt(Validator):
  """Validator that Col1 < Col2."""

  def __init__(self, col1: str, col2: str) -> None:
    self.col1 = col1
    self.col2 = col2

  def __class_getitem__(cls, items: tuple[str, str]) -> Lt:
    return cls(items[0], items[1])

  def get_checks(self) -> list[pa.Check]:
    return [
      pa.Check(
        lambda df: (
          np.all(df[self.col1].values < df[self.col2].values)
          if isinstance(df, pd.DataFrame)
          and self.col1 in df.columns
          and self.col2 in df.columns
          else True
        ),
        name=f"{self.col1} < {self.col2}",
        error=f"{self.col1} must be < {self.col2}",
      )
    ]

  def validate(self, data: Any) -> Any:  # noqa: ANN401
    if not isinstance(data, pd.DataFrame):
      raise TypeError("Lt validator requires a pandas DataFrame")
    return super().validate(data)


class MonoUp(Validator):
  """Validator for monotonically increasing values or index."""

  def get_checks(self) -> list[pa.Check]:
    return [
      pa.Check(
        lambda x: x.is_monotonic_increasing
        if isinstance(x, (pd.Series, pd.Index))
        else x.apply(lambda col: col.is_monotonic_increasing).all(),
        error="must be monotonically increasing",
      )
    ]


class MonoDown(Validator):
  """Validator for monotonically decreasing values or index."""

  def get_checks(self) -> list[pa.Check]:
    return [
      pa.Check(
        lambda x: x.is_monotonic_decreasing
        if isinstance(x, (pd.Series, pd.Index))
        else x.apply(lambda col: col.is_monotonic_decreasing).all(),
        error="must be monotonically decreasing",
      )
    ]


class Index(Validator):
  """Validator for index properties."""

  def __init__(self, *validators: Validator | type[Validator]) -> None:
    self.validators = validators

  def __class_getitem__(
    cls, items: type[Validator] | tuple[type[Validator], ...]
  ) -> Index:
    if not isinstance(items, tuple):
      items = (items,)
    return cls(*items)

  # Index validation is handled in _build_dataframe_schema / _build_series_schema
  def validate(self, data: Any) -> Any:  # noqa: ANN401
    if isinstance(data, (pd.Series, pd.DataFrame)):
      # Validate index
      # We can use _build_series_schema logic but applied to index
      # Or just use the validators directly on the index converted to Series
      s_index = data.index.to_series().reset_index(drop=True)
      for v_item in self.validators:
        v = _instantiate_validator(v_item)
        if v:
          if isinstance(v, Datetime):
            if not isinstance(data.index, pd.DatetimeIndex):
              raise SchemaError(
                schema=None, data=data, message="Index must be DatetimeIndex"
              )
          else:
            # Run check on index values
            v.validate(s_index)
    return data


class HasColumn(Validator):
  """Wrapper to apply validators to specific DataFrame columns."""

  def __init__(self, column: str, *validators: Validator | type[Validator]) -> None:
    self.column = column
    self.validators = validators

  def __class_getitem__(cls, items: str | tuple[str, ...]) -> HasColumn:
    if isinstance(items, str):
      return cls(items)
    column = items[0]
    validators = items[1:] if len(items) > 1 else ()
    return cls(column, *validators)  # type: ignore[arg-type]

  def validate(self, data: Any) -> Any:  # noqa: ANN401
    if isinstance(data, pd.DataFrame):
      if self.column not in data.columns:
        raise SchemaError(
          schema=None,
          data=data,
          message=f"Column '{self.column}' not found in DataFrame",
        )

      # Extract column as Series and validate
      series = data[self.column]
      for v_item in self.validators:
        v = _instantiate_validator(v_item)
        if v:
          v.validate(series)
    return data


P = ParamSpec("P")
R = TypeVar("R")


def _instantiate_validator(item: Any) -> Validator | None:  # noqa: ANN401
  """Helper to instantiate a validator from a type or instance."""
  if isinstance(item, type) and issubclass(item, Validator):
    return item()
  if isinstance(item, Validator):
    return item
  return None


def _build_dataframe_schema(validators: list[Validator]) -> pa.DataFrameSchema:
  """Build DataFrame schema from validators."""
  checks = []
  columns: dict[str, pa.Column] = {}
  index_checks = []
  index_dtype = None

  # 1. Collect global checks and column-specific checks
  for v in validators:
    if isinstance(v, HasColumns):
      for col in v.columns:
        if col not in columns:
          columns[col] = pa.Column(required=True)
    elif isinstance(v, HasColumn):
      # Column specific validation
      col_checks = []
      for sub_v_item in v.validators:
        sub_v = _instantiate_validator(sub_v_item)
        if sub_v:
          col_checks.extend(sub_v.get_checks())

      # Merge into existing column definition or create new
      if v.column in columns:
        # If column exists, we should ideally merge checks.
        # But pa.Column is immutable-ish.
        # For simplicity in this implementation, we might overwrite or ignore
        # duplicates.
        # A robust implementation would collect all checks per column first.
        pass

  # Re-scanning to build schema properly

  # Global DataFrame checks (e.g. Ge, Le)
  for v in validators:
    if isinstance(v, (Ge, Le, Gt, Lt, Finite, NonNaN, NonNegative, Positive)):
      checks.extend(v.get_checks())
    elif isinstance(v, MonoUp):
      checks.append(
        pa.Check(
          lambda df: df.apply(lambda col: col.is_monotonic_increasing).all(),
          error="Values must be monotonically increasing",
        )
      )
    elif isinstance(v, MonoDown):
      checks.append(
        pa.Check(
          lambda df: df.apply(lambda col: col.is_monotonic_decreasing).all(),
          error="Values must be monotonically decreasing",
        )
      )

  # Column definitions
  column_checks_map: dict[str, list[pa.Check]] = {}
  required_columns = set()

  for v in validators:
    if isinstance(v, HasColumns):
      required_columns.update(v.columns)
    elif isinstance(v, HasColumn):
      required_columns.add(v.column)
      if v.column not in column_checks_map:
        column_checks_map[v.column] = []

      for sub_v_item in v.validators:
        sub_v = _instantiate_validator(sub_v_item)
        if sub_v:
          column_checks_map[v.column].extend(sub_v.get_checks())

  # Construct pa.Column objects
  for col_name in required_columns:
    columns[col_name] = pa.Column(
      checks=column_checks_map.get(col_name, []), required=True
    )

  # Index validation
  for v in validators:
    if isinstance(v, Index):
      for sub_v_item in v.validators:
        sub_v = _instantiate_validator(sub_v_item)
        if sub_v:
          if isinstance(sub_v, Datetime):
            index_dtype = "datetime64[ns]"  # or generic datetime
          else:
            index_checks.extend(sub_v.get_checks())

  # Create Schema
  return pa.DataFrameSchema(
    columns=columns,
    checks=checks,
    index=(
      pa.Index(dtype=index_dtype, checks=index_checks)
      if (index_checks or index_dtype)
      else None
    ),
    coerce=False,
  )


def _build_series_schema(validators: list[Validator]) -> pa.SeriesSchema:
  """Build Series schema from validators."""
  checks = []
  dtype = None

  for v in validators:
    if not isinstance(v, (Index, Datetime)):
      checks.extend(v.get_checks())

  index_checks = []
  index_dtype = None

  for v in validators:
    if isinstance(v, Datetime):
      index_dtype = "datetime64[ns]"
    elif isinstance(v, Index):
      for sub_v_item in v.validators:
        sub_v = _instantiate_validator(sub_v_item)
        if sub_v:
          if isinstance(sub_v, Datetime):
            index_dtype = "datetime64[ns]"
          else:
            index_checks.extend(sub_v.get_checks())

  return pa.SeriesSchema(
    dtype=dtype,
    checks=checks,
    index=(
      pa.Index(dtype=index_dtype, checks=index_checks)
      if (index_checks or index_dtype)
      else None
    ),
  )


def validated(func: Callable[P, R]) -> Callable[P, R]:  # noqa: UP047
  """Decorator to validate function arguments based on Annotated types.

  The decorator automatically adds a `skip_validation` parameter to the function.
  When `skip_validation=False` (default), validation is performed. When
  `skip_validation=True`, validation is skipped for maximum performance.

  Args:
    func: The function to decorate.

  Returns:
    The decorated function with automatic validation support.
  """
  # 1. Compute metadata ONCE at decoration time
  sig = inspect.signature(func)

  # Resolve type hints to handle string annotations
  try:
    type_hints = typing.get_type_hints(func, include_extras=True)
  except Exception:
    type_hints = {}

  # Pre-compute a mapping of arg_name -> pa.Schema
  arg_schemas: dict[str, pa.DataFrameSchema | pa.SeriesSchema] = {}

  for name, param in sig.parameters.items():
    if name == "self":
      continue

    annotation = type_hints.get(name, param.annotation)
    origin = get_origin(annotation)

    # Handle Optional[Annotated[...]] / Union[Annotated[...], None]
    if origin is typing.Union:
      args = get_args(annotation)
      for arg in args:
        if get_origin(arg) is Annotated:
          annotation = arg
          origin = Annotated
          break

    if origin is Annotated:
      metadata = get_args(annotation)
      validators = []
      # Iterate over metadata (skipping the first one which is the type)
      for item in metadata[1:]:
        v = _instantiate_validator(item)
        if v:
          validators.append(v)

      if validators:
        # Determine if we should build DataFrame or Series schema
        # We can look at the type hint (metadata[0])
        arg_type = metadata[0]
        if isinstance(arg_type, type) and issubclass(arg_type, pd.DataFrame):
          arg_schemas[name] = _build_dataframe_schema(validators)
        elif isinstance(arg_type, type) and issubclass(arg_type, pd.Series):
          arg_schemas[name] = _build_series_schema(validators)
        else:
          # Fallback: try to build both or decide at runtime?
          # If we don't know the type, we can't pre-build the exact schema type easily
          # unless we assume one.
          # But wait, Annotated[pd.DataFrame, ...] gives us the type!
          # If the user used Annotated[Any, ...], we might have trouble.
          # Let's assume standard usage.
          pass

  @functools.wraps(func)
  def wrapper(*args: P.args, skip_validation: bool = False, **kwargs: P.kwargs) -> R:
    if skip_validation:
      return func(*args, **kwargs)

    # 2. Bind args
    bound = sig.bind(*args, **kwargs)
    bound.apply_defaults()

    # 3. Validate using pre-computed schemas
    for name, value in bound.arguments.items():
      if value is None:
        continue

      if name in arg_schemas:
        try:
          arg_schemas[name].validate(value)
        except Exception as e:
          raise e

    return func(*bound.args, **bound.kwargs)

  return wrapper  # type: ignore[return-value]



"""Tests for individual validator classes."""
# pyright: reportUnknownMemberType=false, reportUnknownVariableType=false
# pyright: reportCallIssue=false, reportAttributeAccessIssue=false

import numpy as np
import pandas as pd
import pytest
from pandera.errors import SchemaError

from pdval import (
  Datetime,
  Finite,
  Ge,
  Gt,
  HasColumn,
  HasColumns,
  Index,
  Le,
  Lt,
  MonoDown,
  MonoUp,
  NonNaN,
  NonNegative,
  Positive,
  Validated,
  validated,
)


class TestFinite:
  """Tests for Finite validator."""

  def test_valid_series(self):
    """Test Finite validator with valid Series."""
    data = pd.Series([1.0, 2.0, 3.0])
    validator = Finite()
    result = validator.validate(data)
    assert result.equals(data)

  def test_valid_dataframe(self):
    """Test Finite validator with valid DataFrame."""
    data = pd.DataFrame({"a": [1.0, 2.0], "b": [3.0, 4.0]})
    validator = Finite()
    result = validator.validate(data)
    assert result.equals(data)

  def test_series_with_inf(self):
    """Test Finite validator rejects Inf."""
    data = pd.Series([1.0, np.inf, 3.0])
    validator = Finite()
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_series_with_nan(self):
    """Test Finite validator rejects NaN."""
    data = pd.Series([1.0, np.nan, 3.0])
    validator = Finite()
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_dataframe_with_inf(self):
    """Test Finite validator rejects DataFrame with Inf."""
    data = pd.DataFrame({"a": [1.0, np.inf], "b": [3.0, 4.0]})
    validator = Finite()
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_empty_series(self):
    """Test Finite validator with empty Series."""
    data = pd.Series([], dtype=float)
    validator = Finite()
    result = validator.validate(data)
    assert len(result) == 0

  def test_non_pandas_type(self):
    """Test Finite validator with non-pandas type."""
    validator = Finite()
    result = validator.validate(42)
    assert result == 42


class TestNonNaN:
  """Tests for NonNaN validator."""

  def test_valid_series(self):
    """Test NonNaN validator with valid Series."""
    data = pd.Series([1.0, 2.0, 3.0])
    validator = NonNaN()
    result = validator.validate(data)
    assert result.equals(data)

  def test_series_with_nan(self):
    """Test NonNaN validator rejects NaN."""
    data = pd.Series([1.0, np.nan, 3.0])
    validator = NonNaN()
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_series_with_inf_allowed(self):
    """Test NonNaN validator allows Inf."""
    data = pd.Series([1.0, np.inf, 3.0])
    validator = NonNaN()
    result = validator.validate(data)
    assert result.equals(data)

  def test_empty_series(self):
    """Test NonNaN validator with empty Series."""
    data = pd.Series([], dtype=float)
    validator = NonNaN()
    result = validator.validate(data)
    assert len(result) == 0


class TestNonNegative:
  """Tests for NonNegative validator."""

  def test_valid_series(self):
    """Test NonNegative validator with valid Series."""
    data = pd.Series([0.0, 1.0, 2.0])
    validator = NonNegative()
    result = validator.validate(data)
    assert result.equals(data)

  def test_series_with_negative(self):
    """Test NonNegative validator rejects negative values."""
    data = pd.Series([1.0, -1.0, 3.0])
    validator = NonNegative()
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_zero_allowed(self):
    """Test NonNegative validator allows zero."""
    data = pd.Series([0.0, 0.0, 0.0])
    validator = NonNegative()
    result = validator.validate(data)
    assert result.equals(data)

  def test_dataframe(self):
    """Test NonNegative validator with DataFrame."""
    data = pd.DataFrame({"a": [1.0, 2.0], "b": [3.0, 4.0]})
    validator = NonNegative()
    result = validator.validate(data)
    assert result.equals(data)


class TestPositive:
  """Tests for Positive validator."""

  def test_valid_series(self):
    """Test Positive validator with valid Series."""
    data = pd.Series([1.0, 2.0, 3.0])
    validator = Positive()
    result = validator.validate(data)
    assert result.equals(data)

  def test_series_with_zero(self):
    """Test Positive validator rejects zero."""
    data = pd.Series([1.0, 0.0, 3.0])
    validator = Positive()
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_series_with_negative(self):
    """Test Positive validator rejects negative values."""
    data = pd.Series([1.0, -1.0, 3.0])
    validator = Positive()
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_all_positive(self):
    """Test Positive validator with all positive values."""
    data = pd.Series([0.1, 100.0, 0.001])
    validator = Positive()
    result = validator.validate(data)
    assert result.equals(data)


class TestIndexDatetime:
  """Tests for Index[Datetime] validator."""

  def test_valid_datetime_index(self):
    """Test Index[Datetime] validator with valid DatetimeIndex."""
    dates = pd.date_range("2024-01-01", periods=3)
    data = pd.Series([1, 2, 3], index=dates)
    validator = Index[Datetime]
    result = validator.validate(data)
    assert result.equals(data)

  def test_invalid_int_index(self):
    """Test Index[Datetime] validator rejects integer index."""
    data = pd.Series([1, 2, 3])
    validator = Index[Datetime]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_dataframe_with_datetime_index(self):
    """Test Index[Datetime] validator with DataFrame."""
    dates = pd.date_range("2024-01-01", periods=3)
    data = pd.DataFrame({"a": [1, 2, 3]}, index=dates)
    validator = Index[Datetime]
    result = validator.validate(data)
    assert result.equals(data)

  def test_non_pandas_type(self):
    """Test Index[Datetime] validator with non-pandas type."""
    validator = Index[Datetime]
    result = validator.validate([1, 2, 3])
    assert result == [1, 2, 3]


class TestIndexMonoUp:
  """Tests for Index[MonoUp] validator."""

  def test_valid_monotonic_increasing(self):
    """Test Index[MonoUp] validator with monotonic index."""
    data = pd.Series([1, 2, 3], index=[0, 1, 2])
    validator = Index[MonoUp]
    result = validator.validate(data)
    assert result.equals(data)

  def test_non_monotonic_index(self):
    """Test Index[MonoUp] validator rejects non-monotonic index."""
    data = pd.Series([1, 2, 3], index=[0, 2, 1])
    validator = Index[MonoUp]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_datetime_monotonic(self):
    """Test Index[MonoUp] validator with datetime index."""
    dates = pd.date_range("2024-01-01", periods=3)
    data = pd.Series([1, 2, 3], index=dates)
    validator = Index[MonoUp]
    result = validator.validate(data)
    assert result.equals(data)

  def test_datetime_non_monotonic(self):
    """Test Index[MonoUp] validator rejects non-monotonic datetime."""
    dates = [
      pd.Timestamp("2024-01-01"),
      pd.Timestamp("2024-01-03"),
      pd.Timestamp("2024-01-02"),
    ]
    data = pd.Series([1, 2, 3], index=dates)
    validator = Index[MonoUp]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)


class TestHasColumns:
  """Tests for HasColumns validator."""

  def test_valid_single_column(self):
    """Test HasColumns validator with single column."""
    data = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    validator = HasColumns["a"]
    result = validator.validate(data)
    assert result.equals(data)

  def test_valid_multiple_columns(self):
    """Test HasColumns validator with multiple columns."""
    data = pd.DataFrame({"a": [1, 2], "b": [3, 4], "c": [5, 6]})
    validator = HasColumns["a", "b"]
    result = validator.validate(data)
    assert result.equals(data)

  def test_missing_single_column(self):
    """Test HasColumns validator with missing column."""
    data = pd.DataFrame({"a": [1, 2]})
    validator = HasColumns["b"]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_missing_multiple_columns(self):
    """Test HasColumns validator with missing columns."""
    data = pd.DataFrame({"a": [1, 2]})
    validator = HasColumns["b", "c"]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_non_dataframe(self):
    """Test HasColumns validator with non-DataFrame."""
    data = pd.Series([1, 2, 3])
    validator = HasColumns["a"]
    with pytest.raises(TypeError, match="requires a pandas DataFrame"):
      validator.validate(data)

  def test_class_getitem_single(self):
    """Test HasColumns __class_getitem__ with single column."""
    validator = HasColumns["col1"]
    assert validator.columns == ["col1"]

  def test_class_getitem_multiple(self):
    """Test HasColumns __class_getitem__ with multiple columns."""
    validator = HasColumns["col1", "col2", "col3"]
    assert validator.columns == ["col1", "col2", "col3"]


class TestGe:
  """Tests for Ge (column comparison) validator."""

  def test_valid_comparison(self):
    """Test Ge validator with valid column comparison."""
    data = pd.DataFrame({"high": [10, 20, 30], "low": [5, 10, 15]})
    validator = Ge["high", "low"]
    result = validator.validate(data)
    assert result.equals(data)

  def test_equal_values_allowed(self):
    """Test Ge validator allows equal values."""
    data = pd.DataFrame({"high": [10, 10, 10], "low": [10, 10, 10]})
    validator = Ge["high", "low"]
    result = validator.validate(data)
    assert result.equals(data)

  def test_invalid_comparison(self):
    """Test Ge validator rejects invalid comparison."""
    data = pd.DataFrame({"high": [10, 5, 30], "low": [5, 10, 15]})
    validator = Ge["high", "low"]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_missing_columns(self):
    """Test Ge validator with missing columns."""
    data = pd.DataFrame({"high": [10, 20]})
    validator = Ge["high", "low"]
    # Should not raise if column is missing
    result = validator.validate(data)
    assert result.equals(data)

  def test_class_getitem(self):
    """Test Ge __class_getitem__."""
    validator = Ge["col1", "col2"]
    assert validator.col1 == "col1"
    assert validator.col2 == "col2"

  def test_non_dataframe(self):
    """Test Ge validator with non-DataFrame."""
    data = pd.Series([1, 2, 3])
    validator = Ge["a", "b"]
    with pytest.raises(TypeError, match="requires a pandas DataFrame"):
      validator.validate(data)


class TestLe:
  """Tests for Le (<=) validator."""

  def test_valid_comparison(self):
    """Test Le validator with valid column comparison."""
    data = pd.DataFrame({"low": [5, 10, 15], "high": [10, 20, 30]})
    validator = Le["low", "high"]
    result = validator.validate(data)
    assert result.equals(data)

  def test_equal_values_allowed(self):
    """Test Le validator allows equal values."""
    data = pd.DataFrame({"low": [10, 10, 10], "high": [10, 10, 10]})
    validator = Le["low", "high"]
    result = validator.validate(data)
    assert result.equals(data)

  def test_invalid_comparison(self):
    """Test Le validator rejects invalid comparison."""
    data = pd.DataFrame({"low": [10, 25, 15], "high": [10, 20, 30]})
    validator = Le["low", "high"]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_class_getitem(self):
    """Test Le __class_getitem__."""
    validator = Le["col1", "col2"]
    assert validator.col1 == "col1"
    assert validator.col2 == "col2"


class TestGt:
  """Tests for Gt (>) validator."""

  def test_valid_comparison(self):
    """Test Gt validator with valid column comparison."""
    data = pd.DataFrame({"high": [20, 30, 40], "low": [10, 20, 30]})
    validator = Gt["high", "low"]
    result = validator.validate(data)
    assert result.equals(data)

  def test_equal_values_rejected(self):
    """Test Gt validator rejects equal values."""
    data = pd.DataFrame({"high": [10, 10, 10], "low": [10, 10, 10]})
    validator = Gt["high", "low"]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_invalid_comparison(self):
    """Test Gt validator rejects invalid comparison."""
    data = pd.DataFrame({"high": [10, 15, 30], "low": [10, 20, 15]})
    validator = Gt["high", "low"]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_class_getitem(self):
    """Test Gt __class_getitem__."""
    validator = Gt["col1", "col2"]
    assert validator.col1 == "col1"
    assert validator.col2 == "col2"


class TestLt:
  """Tests for Lt (<) validator."""

  def test_valid_comparison(self):
    """Test Lt validator with valid column comparison."""
    data = pd.DataFrame({"low": [10, 20, 30], "high": [20, 30, 40]})
    validator = Lt["low", "high"]
    result = validator.validate(data)
    assert result.equals(data)

  def test_equal_values_rejected(self):
    """Test Lt validator rejects equal values."""
    data = pd.DataFrame({"low": [10, 10, 10], "high": [10, 10, 10]})
    validator = Lt["low", "high"]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_invalid_comparison(self):
    """Test Lt validator rejects invalid comparison."""
    data = pd.DataFrame({"low": [15, 20, 30], "high": [10, 30, 25]})
    validator = Lt["low", "high"]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_class_getitem(self):
    """Test Lt __class_getitem__."""
    validator = Lt["col1", "col2"]
    assert validator.col1 == "col1"
    assert validator.col2 == "col2"


class TestMonoUp:
  """Tests for MonoUp (monotonically increasing) validator."""

  def test_valid_series_increasing(self):
    """Test MonoUp validator with valid increasing Series."""
    data = pd.Series([1, 2, 3, 4, 5])
    validator = MonoUp()
    result = validator.validate(data)
    assert result.equals(data)

  def test_valid_series_equal(self):
    """Test MonoUp validator allows equal consecutive values."""
    data = pd.Series([1, 2, 2, 3, 3, 4])
    validator = MonoUp()
    result = validator.validate(data)
    assert result.equals(data)

  def test_invalid_series_decreasing(self):
    """Test MonoUp validator rejects decreasing values."""
    data = pd.Series([1, 2, 3, 2, 5])
    validator = MonoUp()
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_valid_dataframe(self):
    """Test MonoUp validator with DataFrame."""
    data = pd.DataFrame({"a": [1, 2, 3], "b": [10, 20, 30]})
    validator = MonoUp()
    result = validator.validate(data)
    assert result.equals(data)

  def test_invalid_dataframe(self):
    """Test MonoUp validator rejects DataFrame with non-monotonic column."""
    data = pd.DataFrame({"a": [1, 2, 3], "b": [10, 5, 30]})
    validator = MonoUp()
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)


class TestMonoDown:
  """Tests for MonoDown (monotonically decreasing) validator."""

  def test_valid_series_decreasing(self):
    """Test MonoDown validator with valid decreasing Series."""
    data = pd.Series([5, 4, 3, 2, 1])
    validator = MonoDown()
    result = validator.validate(data)
    assert result.equals(data)

  def test_valid_series_equal(self):
    """Test MonoDown validator allows equal consecutive values."""
    data = pd.Series([5, 4, 4, 3, 3, 2])
    validator = MonoDown()
    result = validator.validate(data)
    assert result.equals(data)

  def test_invalid_series_increasing(self):
    """Test MonoDown validator rejects increasing values."""
    data = pd.Series([5, 4, 3, 4, 1])
    validator = MonoDown()
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_valid_dataframe(self):
    """Test MonoDown validator with DataFrame."""
    data = pd.DataFrame({"a": [3, 2, 1], "b": [30, 20, 10]})
    validator = MonoDown()
    result = validator.validate(data)
    assert result.equals(data)

  def test_invalid_dataframe(self):
    """Test MonoDown validator rejects DataFrame with non-monotonic column."""
    data = pd.DataFrame({"a": [3, 2, 1], "b": [30, 35, 10]})
    validator = MonoDown()
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)


class TestHasColumn:
  """Tests for HasColumn wrapper validator."""

  def test_single_validator(self):
    """Test HasColumn with single validator."""
    data = pd.DataFrame({"a": [1.0, 2.0, 3.0], "b": [4.0, 5.0, 6.0]})
    validator = HasColumn["a", Finite]
    result = validator.validate(data)
    assert result.equals(data)

  def test_single_validator_fails(self):
    """Test HasColumn validator fails when column violates constraint."""
    data = pd.DataFrame({"a": [1.0, np.inf, 3.0], "b": [4.0, 5.0, 6.0]})
    validator = HasColumn["a", Finite]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_multiple_validators(self):
    """Test HasColumn with multiple validators."""
    data = pd.DataFrame({"a": [1.0, 2.0, 3.0], "b": [4.0, 5.0, 6.0]})
    validator = HasColumn["a", Finite, Positive]
    result = validator.validate(data)
    assert result.equals(data)

  def test_multiple_validators_fails(self):
    """Test HasColumn with multiple validators where one fails."""
    data = pd.DataFrame({"a": [1.0, 0.0, 3.0], "b": [4.0, 5.0, 6.0]})
    validator = HasColumn["a", Finite, Positive]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_missing_column(self):
    """Test HasColumn with missing column."""
    data = pd.DataFrame({"b": [1.0, 2.0, 3.0]})
    validator = HasColumn["a", Finite]
    with pytest.raises(SchemaError, match=None):
      validator.validate(data)

  def test_monotonic_validator(self):
    """Test HasColumn with MonoUp validator."""
    data = pd.DataFrame({"a": [1, 2, 3], "b": [10, 5, 3]})
    validator = HasColumn["a", MonoUp]
    result = validator.validate(data)
    assert result.equals(data)

    # Column b is not monotonic up
    validator_b = HasColumn["b", MonoUp]
    with pytest.raises(SchemaError, match=None):
      validator_b.validate(data)

  def test_class_getitem(self):
    """Test HasColumn __class_getitem__."""
    validator = HasColumn["col", Finite, Positive]
    assert validator.column == "col"
    assert len(validator.validators) == 2

  def test_class_getitem_no_validators(self):
    """Test HasColumn __class_getitem__ with no validators."""
    # HasColumn["col"] just checks column existence
    validator = HasColumn["col"]
    assert validator.column == "col"
    assert len(validator.validators) == 0

  def test_column_presence_only(self):
    """Test HasColumn just checks column presence when no validators."""
    data = pd.DataFrame({"a": [1.0, np.inf, -5.0], "b": [4.0, 5.0, 6.0]})

    # Should pass - column exists (even with invalid values)
    validator = HasColumn["a"]
    result = validator.validate(data)
    assert result.equals(data)

    # Should fail - column doesn't exist
    validator_missing = HasColumn["missing"]
    with pytest.raises(SchemaError, match=None):
      validator_missing.validate(data)


class TestHasColumnWithDecorator:
  """Tests for HasColumn used with @validated decorator."""

  def test_single_column_validation(self):
    """Test @validated with HasColumn for single column."""

    @validated
    def process(
      data: Validated[pd.DataFrame, HasColumn["price", Finite, Positive]],  # noqa: F821
      validate: bool = True,
    ):
      return data["price"].sum()

    valid_data = pd.DataFrame({"price": [100.0, 200.0, 150.0], "volume": [10, 20, 15]})
    result = process(valid_data)
    assert result == 450.0

    # Fails Finite check
    invalid_data = pd.DataFrame(
      {
        "price": [100.0, np.inf, 150.0],
        "volume": [10, 20, 15],
      }
    )
    with pytest.raises(SchemaError, match=None):
      process(invalid_data)

    # Fails Positive check
    invalid_data = pd.DataFrame({"price": [100.0, 0.0, 150.0], "volume": [10, 20, 15]})
    with pytest.raises(SchemaError, match=None):
      process(invalid_data)

  def test_multiple_column_validation(self):
    """Test @validated with multiple HasColumn validators."""

    @validated
    def process(
      data: Validated[
        pd.DataFrame,
        HasColumn["price", Finite, Positive],  # noqa: F821
        HasColumn["volume", Positive],  # noqa: F821
      ],
    ):
      return (data["price"] * data["volume"]).sum()

    valid_data = pd.DataFrame({"price": [100.0, 200.0, 150.0], "volume": [10, 20, 15]})
    result = process(valid_data)
    assert result == 100 * 10 + 200 * 20 + 150 * 15

    # Fails volume Positive check
    invalid_data = pd.DataFrame(
      {
        "price": [100.0, 200.0, 150.0],
        "volume": [10, -5, 15],
      }
    )
    with pytest.raises(SchemaError, match=None):
      process(invalid_data)

  def test_oncolumn_with_monotonic(self):
    """Test HasColumn with MonoUp validator."""

    @validated
    def process(
      data: Validated[pd.DataFrame, HasColumn["timestamp", MonoUp]],  # noqa: F821
      validate: bool = True,
    ):
      return len(data)

    valid_data = pd.DataFrame(
      {
        "timestamp": [1, 2, 3, 4, 5],
        "value": [10, 20, 30, 40, 50],
      }
    )
    result = process(valid_data)
    assert result == 5

    # Fails MonoUp check
    invalid_data = pd.DataFrame(
      {
        "timestamp": [1, 2, 5, 4, 3],
        "value": [10, 20, 30, 40, 50],
      }
    )
    with pytest.raises(SchemaError, match=None):
      process(invalid_data)


class TestCombinedValidators:
  """Tests combining new validators with existing ones."""

  def test_comparison_validators_combined(self):
    """Test using multiple comparison validators."""

    @validated
    def process(
      data: Validated[pd.DataFrame, Gt["high", "low"], Le["low", "close"]],  # noqa: F821
      validate: bool = True,
    ):
      return len(data)

    valid_data = pd.DataFrame(
      {
        "high": [105, 110, 108],
        "low": [100, 105, 103],
        "close": [102, 107, 105],
      }
    )
    result = process(valid_data)
    assert result == 3

    # Fails Gt check (high not > low)
    invalid_data = pd.DataFrame(
      {
        "high": [100, 110, 108],
        "low": [100, 105, 103],
        "close": [102, 107, 105],
      }
    )
    with pytest.raises(SchemaError, match=None):
      process(invalid_data)

  def test_oncolumn_with_nonnan(self):
    """Test HasColumn with NonNaN validator."""

    @validated
    def process(
      data: Validated[pd.DataFrame, HasColumn["price", NonNaN, Positive]],  # noqa: F821
      validate: bool = True,
    ):
      return data["price"].mean()

    valid_data = pd.DataFrame({"price": [100.0, 200.0, 150.0]})
    result = process(valid_data)
    assert result == 150.0

    invalid_data = pd.DataFrame({"price": [100.0, np.nan, 150.0]})
    with pytest.raises(SchemaError, match=None):
      process(invalid_data)



"""Tests for the @validated decorator."""
# pyright: reportUnknownMemberType=false, reportUnknownVariableType=false
# pyright: reportCallIssue=false

import numpy as np
import pandas as pd
import pytest
from pandera.errors import SchemaError

from pdval import (
  Datetime,
  Finite,
  Ge,
  HasColumns,
  Index,
  MonoUp,
  Positive,
  Validated,
  validated,
)


class TestValidatedDecorator:
  """Tests for @validated decorator basic functionality."""

  def test_function_with_validation(self):
    """Test @validated decorator validates arguments."""

    @validated
    def process(data: Validated[pd.Series, Finite]):
      return data.sum()

    valid_data = pd.Series([1.0, 2.0, 3.0])
    result = process(valid_data)
    assert result == 6.0

  def test_function_rejects_invalid_data(self):
    """Test @validated decorator rejects invalid data."""

    @validated
    def process(data: Validated[pd.Series, Finite]):
      return data.sum()

    invalid_data = pd.Series([1.0, np.inf, 3.0])
    with pytest.raises(SchemaError, match=None):
      process(invalid_data)

  def test_validation_can_be_disabled(self):
    """Test validation can be disabled with skip_validation=True."""

    @validated
    def process(data: Validated[pd.Series, Finite]):
      return data.sum()

    invalid_data = pd.Series([1.0, np.inf, 3.0])
    # Should not raise when validation is disabled
    result = process(invalid_data, skip_validation=True)
    assert np.isinf(result)

  def test_multiple_validators(self):
    """Test multiple validators in chain."""

    @validated
    def process(data: Validated[pd.Series, Finite, Positive]):
      return data.sum()

    # Valid data
    valid_data = pd.Series([1.0, 2.0, 3.0])
    result = process(valid_data)
    assert result == 6.0

    # Fails Finite check
    with pytest.raises(SchemaError, match=None):
      process(pd.Series([1.0, np.nan, 3.0]))

    # Fails Positive check
    with pytest.raises(SchemaError, match=None):
      process(pd.Series([1.0, 0.0, 3.0]))

  def test_dataframe_validation(self):
    """Test DataFrame validation."""

    @validated
    def process(
      data: Validated[pd.DataFrame, HasColumns["a", "b"], Finite],  # noqa: F821
      validate: bool = True,
    ):
      return data["a"] + data["b"]

    valid_data = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    result = process(valid_data)
    assert result.tolist() == [4, 6]

    # Missing column
    with pytest.raises(SchemaError, match=None):
      process(pd.DataFrame({"a": [1, 2]}))

  def test_preserves_function_metadata(self):
    """Test decorator preserves function name and docstring."""

    @validated
    def my_function(data: Validated[pd.Series, Finite]):
      """My docstring."""
      return data.sum()

    assert my_function.__name__ == "my_function"
    assert my_function.__doc__ == "My docstring."

  def test_works_with_methods(self):
    """Test @validated works with class methods."""

    class Processor:
      @validated
      def process(self, data: Validated[pd.Series, Finite]):
        return data.sum()

    processor = Processor()
    valid_data = pd.Series([1.0, 2.0, 3.0])
    result = processor.process(valid_data)
    assert result == 6.0

  def test_optional_validated_argument(self):
    """Test Optional[Validated[...]] annotation."""

    @validated
    def process(data: Validated[pd.Series, Finite] | None = None):
      if data is None:
        return 0
      return data.sum()

    # None is allowed
    result = process(None)
    assert result == 0

    # Valid data works
    result = process(pd.Series([1.0, 2.0, 3.0]))
    assert result == 6.0

    # Invalid data still raises
    with pytest.raises(SchemaError, match=None):
      process(pd.Series([1.0, np.inf, 3.0]))

  def test_multiple_arguments(self):
    """Test validation with multiple arguments."""

    @validated
    def combine(
      data1: Validated[pd.Series, Finite],
      data2: Validated[pd.Series, Finite],
    ):
      return data1 + data2

    valid1 = pd.Series([1.0, 2.0])
    valid2 = pd.Series([3.0, 4.0])
    result = combine(valid1, valid2)
    assert result.tolist() == [4.0, 6.0]

    # First argument invalid
    with pytest.raises(SchemaError, match=None):
      combine(pd.Series([np.inf, 2.0]), valid2)

    # Second argument invalid
    with pytest.raises(SchemaError, match=None):
      combine(valid1, pd.Series([3.0, np.nan]))

  def test_non_validated_arguments_ignored(self):
    """Test non-validated arguments are not validated."""

    @validated
    def process(
      data: Validated[pd.Series, Finite],
      multiplier: float,
    ):
      return data * multiplier

    valid_data = pd.Series([1.0, 2.0, 3.0])
    result = process(valid_data, multiplier=2.0)
    assert result.tolist() == [2.0, 4.0, 6.0]


class TestComplexValidations:
  """Tests for complex validation scenarios."""

  def test_ohlc_validation(self):
    """Test validation of OHLC data."""

    @validated
    def calculate_true_range(
      data: Validated[
        pd.DataFrame, HasColumns["high", "low", "close"], Ge["high", "low"]  # noqa: F821
      ],
    ):
      hl = data["high"] - data["low"]
      hc = abs(data["high"] - data["close"].shift(1))
      lc = abs(data["low"] - data["close"].shift(1))
      return pd.concat([hl, hc, lc], axis=1).max(axis=1)

    # Valid OHLC
    valid_data = pd.DataFrame(
      {
        "high": [102, 105, 104],
        "low": [100, 103, 101],
        "close": [101, 104, 102],
      }
    )
    result = calculate_true_range(valid_data)
    assert len(result) == 3

    # High < Low should fail
    invalid_data = pd.DataFrame(
      {
        "high": [100, 105, 104],
        "low": [102, 103, 101],
        "close": [101, 104, 102],
      }
    )
    with pytest.raises(SchemaError, match=None):
      calculate_true_range(invalid_data)

  def test_time_series_validation(self):
    """Test time series specific validation."""

    @validated
    def resample_data(
      data: Validated[pd.Series, Index[Datetime, MonoUp], Finite],
      freq: str = "1D",
    ):
      return data.resample(freq).mean()

    # Valid time series
    dates = pd.date_range("2024-01-01", periods=10, freq="h")
    valid_data = pd.Series(range(10), index=dates)
    result = resample_data(valid_data)
    assert isinstance(result.index, pd.DatetimeIndex)

    # Non-datetime index
    invalid_data = pd.Series(range(10))
    with pytest.raises(SchemaError, match=None):
      resample_data(invalid_data)

    # Non-monotonic datetime index
    dates_shuffled = [
      pd.Timestamp("2024-01-01"),
      pd.Timestamp("2024-01-03"),
      pd.Timestamp("2024-01-02"),
    ]
    non_monotonic = pd.Series([1, 2, 3], index=dates_shuffled)
    with pytest.raises(SchemaError, match=None):
      resample_data(non_monotonic)

  def test_percentage_returns_validation(self):
    """Test validation for percentage returns calculation."""

    @validated
    def calculate_returns(prices: Validated[pd.Series, Finite, Positive]):
      return prices.pct_change()

    # Valid prices
    valid_prices = pd.Series([100.0, 102.0, 101.0, 103.0])
    result = calculate_returns(valid_prices)
    assert len(result) == 4

    # Zero price fails Positive check
    with pytest.raises(SchemaError, match=None):
      calculate_returns(pd.Series([100.0, 0.0, 101.0]))

    # NaN price fails Finite check
    with pytest.raises(SchemaError, match=None):
      calculate_returns(pd.Series([100.0, np.nan, 101.0]))


class TestEdgeCases:
  """Tests for edge cases and error conditions."""

  def test_empty_series(self):
    """Test validation with empty Series."""

    @validated
    def process(data: Validated[pd.Series, Finite]):
      return len(data)

    empty_data = pd.Series([], dtype=float)
    result = process(empty_data)
    assert result == 0

  def test_validator_class_vs_instance(self):
    """Test both Validator class and instance work."""

    @validated
    def with_class(data: Validated[pd.Series, Finite]):
      return data.sum()

    @validated
    def with_instance(data: Validated[pd.Series, Finite()]):
      return data.sum()

    valid_data = pd.Series([1.0, 2.0, 3.0])
    invalid_data = pd.Series([1.0, np.inf, 3.0])

    # Both should work with valid data
    assert with_class(valid_data) == 6.0
    assert with_instance(valid_data) == 6.0

    # Both should reject invalid data
    with pytest.raises(SchemaError):
      with_class(invalid_data)
    with pytest.raises(SchemaError):
      with_instance(invalid_data)

  def test_function_without_validate_param(self):
    """Test function without validate parameter defaults to True."""

    @validated
    def process(data: Validated[pd.Series, Finite]):
      return data.sum()

    valid_data = pd.Series([1.0, 2.0, 3.0])
    result = process(valid_data)
    assert result == 6.0

    # Should still validate and reject invalid data
    invalid_data = pd.Series([1.0, np.inf, 3.0])
    with pytest.raises(SchemaError, match=None):
      process(invalid_data)

  def test_kwargs_arguments(self):
    """Test validation works with keyword arguments."""

    @validated
    def process(data: Validated[pd.Series, Finite]):
      return data.sum()

    valid_data = pd.Series([1.0, 2.0, 3.0])

    # Positional
    result = process(valid_data)
    assert result == 6.0

    # Keyword
    result = process(data=valid_data)
    assert result == 6.0

    # Mixed
    result = process(valid_data, skip_validation=False)
    assert result == 6.0

  def test_default_argument_values(self):
    """Test validation with default argument values."""

    @validated
    def process(
      data: Validated[pd.Series, Finite] = pd.Series([1.0, 2.0]),
    ):
      return data.sum()

    # No arguments (uses default)
    result = process()
    assert result == 3.0

    # Override default
    result = process(pd.Series([5.0, 6.0]))
    assert result == 11.0
